package ru.dbotthepony.mc.otm.capability

import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import net.minecraft.ChatFormatting
import net.minecraft.core.Direction
import net.minecraftforge.common.capabilities.ICapabilityProvider
import net.minecraftforge.common.util.INBTSerializable
import net.minecraft.nbt.CompoundTag
import net.minecraft.world.item.ItemStack
import ru.dbotthepony.mc.otm.android.AndroidFeatureType
import ru.dbotthepony.mc.otm.android.AndroidFeature
import java.lang.Runnable
import net.minecraft.server.level.ServerPlayer
import net.minecraftforge.event.entity.living.LivingHurtEvent
import net.minecraft.nbt.ListTag
import net.minecraft.nbt.Tag
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.effect.MobEffect
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.fml.event.lifecycle.FMLCommonSetupEvent
import net.minecraft.world.effect.MobEffects
import net.minecraft.world.entity.Entity
import net.minecraft.world.entity.player.Player
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.capabilities.ForgeCapabilities
import net.minecraftforge.event.AttachCapabilitiesEvent
import net.minecraftforge.eventbus.api.SubscribeEvent
import net.minecraftforge.event.entity.living.LivingEvent.LivingTickEvent
import net.minecraftforge.event.entity.player.PlayerEvent
import net.minecraftforge.eventbus.api.EventPriority
import ru.dbotthepony.mc.otm.*
import ru.dbotthepony.mc.otm.android.AndroidResearch
import ru.dbotthepony.mc.otm.android.AndroidResearchType
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.network.*
import ru.dbotthepony.mc.otm.registry.AndroidFeatures
import ru.dbotthepony.mc.otm.registry.MNames
import ru.dbotthepony.mc.otm.registry.MRegistry
import ru.dbotthepony.mc.otm.registry.StatNames
import java.util.*

@Suppress("unused")
class MatteryPlayerCapability(val ply: Player) : ICapabilityProvider, IMatteryEnergyStorage, INBTSerializable<CompoundTag> {
	private var battery = ImpreciseFraction.ZERO
	private var maxBattery = ImpreciseFraction(60000)

	private val features = Object2ObjectArrayMap<AndroidFeatureType<*>, AndroidFeature>()
	private val networkQueue = ArrayList<Any>()
	private val queuedTicks = ArrayList<Runnable>()
	private var tickedOnce = false

	private var willBecomeAndroidNetwork = false
	private var shouldPlaySound = false
	private val research = IdentityHashMap<AndroidResearchType<*>, AndroidResearch>()

	private var remoteBattery = ImpreciseFraction(-1)
	private var remoteMaxBattery = ImpreciseFraction(-1)
	private var remoteBatteryStack = ItemStack.EMPTY

	private var invalidateNetworkIn = 10
	private var remoteIsAndroid = false

	var batteryItemStack: ItemStack = ItemStack.EMPTY

	var willBecomeAndroid = false
	var isAndroid = false
	var sleepTicks = 0
	val isEverAndroid: Boolean get() = isAndroid || willBecomeAndroid
	var lastJumpTicks = 14

	fun invalidateNetworkState() {
		invalidateNetworkIn = 10
	}

	fun becomeAndroidSoft() {
		if (isAndroid || willBecomeAndroid) return
		willBecomeAndroid = true
		(ply as? ServerPlayer)?.displayClientMessage(TranslatableComponent("otm.pill.message").withStyle(ChatFormatting.GRAY), false)
	}

	fun becomeAndroid() {
		if (isAndroid) return

		isAndroid = true
		willBecomeAndroid = false
		shouldPlaySound = false
		battery = ImpreciseFraction(60000)
		maxBattery = ImpreciseFraction(60000)
	}

	fun becomeAndroidAndKill() {
		if (isAndroid) return

		becomeAndroid()
		ply.hurt(MRegistry.DAMAGE_BECOME_ANDROID, ply.maxHealth * 2)
	}

	fun becomeHumane() {
		if (willBecomeAndroid) willBecomeAndroid = false
		if (!isAndroid) return

		isAndroid = false
		shouldPlaySound = false
		battery = ImpreciseFraction(0)
		maxBattery = ImpreciseFraction(60000)
		dropBattery()
	}

	fun becomeHumaneAndKill() {
		if (willBecomeAndroid) willBecomeAndroid = false
		if (!isAndroid) return

		becomeHumane()
		ply.hurt(MRegistry.DAMAGE_BECOME_HUMANE, ply.maxHealth * 2)
	}

	fun obliviate(refund: Boolean = true) {
		if (refund) {
			for (instance in research.values) {
				if (instance.isResearched) {
					instance.unResearch()
					instance.refund(simulate = false)
				}
			}
		}

		val copy = java.util.List.copyOf(features.values)

		for (feature in copy) {
			removeFeature(feature.type)
		}
	}

	fun <T : AndroidResearch> getResearch(type: AndroidResearchType<T>): T {
		return research.computeIfAbsent(type) {
			return@computeIfAbsent type.factory(this)
		} as T
	}

	private fun addFeature(feature: AndroidFeature): Boolean {
		if (features.containsKey(feature.type)) return false
		features[feature.type] = feature

		if (!ply.level.isClientSide) {
			queuedTicks.add(feature::applyModifiers)
		}

		if (ply is ServerPlayer) {
			sendNetwork(AndroidFeatureSyncPacket(feature))
		}

		return true
	}

	@Suppress("unchecked_cast")
	fun <T : AndroidFeature> addFeature(feature: AndroidFeatureType<T>): T {
		val get = features[feature]
		if (get != null) return get as T

		val factory = feature.create(this)

		features[feature] = factory

		if (!ply.level.isClientSide) {
			queuedTicks.add(factory::applyModifiers)
		}

		if (ply is ServerPlayer) {
			sendNetwork(AndroidFeatureSyncPacket(factory))
		}

		return factory
	}

	fun removeFeature(feature: AndroidFeatureType<*>): Boolean {
		val removed = features.remove(feature)

		if (removed != null) {
			if (!ply.level.isClientSide) {
				queuedTicks.add {
					removed.removeModifiers()
				}
			}

			if (ply is ServerPlayer) {
				sendNetwork(AndroidFeatureRemovePacket(removed.type))
			}
		}

		return removed != null
	}

	fun hasFeature(feature: AndroidFeatureType<*>): Boolean {
		return features.containsKey(feature)
	}

	fun hasFeatureLevel(feature: AndroidFeatureType<*>, level: Int): Boolean {
		val get = features[feature] ?: return false
		return get.level >= level
	}

	@Suppress("unchecked_cast")
	fun <T : AndroidFeature> getFeature(feature: AndroidFeatureType<T>): T? {
		return features[feature] as T?
	}

	fun onHurt(event: LivingHurtEvent) {
		for (feature in features.values) {
			feature.onHurt(event)
		}
	}

	fun <T : AndroidFeature> computeIfAbsent(feature: AndroidFeatureType<T>): T {
		return getFeature(feature) ?: addFeature(feature)
	}

	fun <T : AndroidFeature> getFeatureO(feature: AndroidFeatureType<T>): Optional<T> {
		val get = getFeature(feature)
		return if (get != null) Optional.of(get) else Optional.empty()
	}

	fun <T : AndroidFeature> ifFeature(feature: AndroidFeatureType<T>, consumer: (T) -> Unit) {
		val get = getFeature(feature)

		if (get != null) {
			consumer(get)
		}
	}

	override fun serializeNBT(): CompoundTag {
		val tag = CompoundTag()

		tag["energy_stored"] = battery.serializeNBT()
		tag["energy_stored_max"] = maxBattery.serializeNBT()
		tag["battery"] = batteryItemStack.serializeNBT()

		val featureList = ListTag()

		for (feature in features.values) {
			val featureNbt = feature.serializeNBT()

			featureNbt["id"] = feature.type.registryName!!.toString()
			featureList.add(featureNbt)
		}

		tag["features"] = featureList

		tag["is_android"] = isAndroid
		tag["will_become_android"] = willBecomeAndroid

		val list = ListTag()

		for ((type, instance) in research) {
			val researchTag = instance.serializeNBT()

			researchTag["id"] = type.registryName!!.toString()
			list.add(researchTag)
		}

		tag["research"] = list

		return tag
	}

	override fun deserializeNBT(compound: CompoundTag) {
		compound.ifHas("energy_stored") {
			battery = ImpreciseFraction.deserializeNBT(it)
		}

		compound.ifHas("energy_stored_max") {
			maxBattery = ImpreciseFraction.deserializeNBT(it)
		}

		compound.ifHas("battery", CompoundTag::class.java) {
			batteryItemStack = ItemStack.of(it)
		}

		features.clear()

		val featuresNbt = compound.getList("features", Tag.TAG_COMPOUND.toInt())

		for (tag in featuresNbt) {
			if (tag is CompoundTag) {
				val feature = MRegistry.ANDROID_FEATURES.getValue(ResourceLocation(tag.getString("id")))

				if (feature?.isApplicable(this) == true) {
					val instance = feature.create(this)

					instance.deserializeNBT(tag)
					addFeature(instance)

					if (!ply.level.isClientSide) {
						queuedTicks.add(instance::applyModifiers)
					}
				}
			}
		}

		isAndroid = compound.getBoolean("is_android")
		willBecomeAndroid = compound.getBoolean("will_become_android")
		research.clear()

		val list = compound.getList("research", Tag.TAG_COMPOUND.toInt())

		for (tag in list) {
			if (tag is CompoundTag) {
				val research = MRegistry.ANDROID_RESEARCH.getValue(ResourceLocation(tag.getString("id")))

				if (research != null) {
					val instance = research.factory(this)
					instance.deserializeNBT(tag)
					this.research[research] = instance
				}
			}
		}
	}

	fun dropBattery() {
		if (batteryItemStack.isEmpty) return
		ply.spawnAtLocation(batteryItemStack)
		batteryItemStack = ItemStack.EMPTY
	}

	private fun sendNetwork(packet: Any) {
		if (ply is ServerPlayer) {
			if (tickedOnce) {
				MatteryPlayerNetworkChannel.send(ply, packet)
			} else {
				networkQueue.add(packet)
			}
		}
	}

	fun tickClient() {
		queuedTicks.clear()

		if (!ply.isAlive) {
			return
		}

		if (willBecomeAndroid) {
			if (ply.isSleeping) {
				sleepTicks++
			} else {
				sleepTicks = 0
			}
		}

		if (isAndroid) {
			for (feature in features.values) {
				feature.tickClient()
			}
		}
	}

	fun tick() {
		if (!ply.isAlive) return

		if (willBecomeAndroid) {
			if (ply.isSleeping) {
				sleepTicks++

				if (sleepTicks > SLEEP_TICKS_LIMIT) {
					becomeAndroid()
					shouldPlaySound = true
					sleepTicks = 0

					(ply as? ServerPlayer)?.displayClientMessage(TranslatableComponent("otm.pill.message_finish").withStyle(ChatFormatting.DARK_RED), false)
				}
			} else {
				sleepTicks = 0
			}
		}

		if (isAndroid) {
			if (ply.airSupply < ply.maxAirSupply)
				ply.airSupply = ply.maxAirSupply

			for (effect in UNAFFECTED_EFFECTS)
				if (ply.hasEffect(effect))
					ply.removeEffect(effect)

			if (!batteryItemStack.isEmpty && battery < maxBattery) {
				batteryItemStack.getCapability(ForgeCapabilities.ENERGY).ifPresent {
					if (it is IMatteryEnergyStorage) {
						battery += it.extractEnergyInner(maxBattery - battery, false)
					} else {
						battery += it.extractEnergy(maxBattery - battery, false)
					}
				}
			}

			// TODO: Maybe passive drain?
			// extractEnergyInner(BigDecimal.valueOf(new Random().nextDouble()), false);
			if (ply.isSwimming && !hasFeature(AndroidFeatures.AIR_BAGS)) {
				ply.isSwimming = false
			}

			val stats = ply.foodData

			while (stats.foodLevel < 18 && extractEnergyInner(ENERGY_FOR_HUNGER_POINT, true) >= ENERGY_FOR_HUNGER_POINT) {
				extractEnergyInner(ENERGY_FOR_HUNGER_POINT, false)
				stats.foodLevel = stats.foodLevel + 1
			}

			// "block" quick regeneration
			// also cause power to generate while in peaceful
			while (stats.foodLevel > 18 && receiveEnergyInner(ENERGY_FOR_HUNGER_POINT, true) >= ENERGY_FOR_HUNGER_POINT) {
				receiveEnergyInner(ENERGY_FOR_HUNGER_POINT, false)
				stats.foodLevel = stats.foodLevel - 1
			}

			val foodLevel = stats.foodLevel.toFloat()

			if (stats.saturationLevel < foodLevel) {
				val extracted = extractEnergyInner(ENERGY_FOR_HUNGER_POINT * (foodLevel - stats.saturationLevel), false)
				stats.setSaturation(stats.saturationLevel + (extracted / ENERGY_FOR_HUNGER_POINT).toFloat())
			}

			if (stats.exhaustionLevel > 0f) {
				val extracted = extractEnergyInner(ENERGY_FOR_HUNGER_POINT * (stats.exhaustionLevel / 4f), false)
				stats.setExhaustion(stats.exhaustionLevel - (extracted / ENERGY_FOR_HUNGER_POINT).toFloat() * 4f)
			}

			for (feature in features.values) {
				feature.tickServer()
			}
		}

		for (runnable in queuedTicks) {
			runnable.run()
		}

		queuedTicks.clear()

		if (invalidateNetworkIn > 0 && --invalidateNetworkIn == 0) {
			remoteBattery = ImpreciseFraction.MINUS_ONE
			remoteMaxBattery = ImpreciseFraction.MINUS_ONE
			remoteBatteryStack = ItemStack.EMPTY

			remoteIsAndroid = false
			willBecomeAndroidNetwork = false

			for (instance in research.values) {
				instance.isDirty = true
				instance.invalidateNetwork()
			}

			for (feature in features.values) {
				feature.isDirty = true
				feature.invalidateNetwork()
			}
		}

		tickedOnce = true

		if (battery != remoteBattery) {
			remoteBattery = battery
			sendNetwork(AndroidEnergyLevelPacket(level = battery, isMaxEnergy = false))
		}

		if (maxBattery != remoteMaxBattery) {
			remoteMaxBattery = maxBattery
			sendNetwork(AndroidEnergyLevelPacket(level = maxBattery, isMaxEnergy = true))
		}

		if (!remoteBatteryStack.equals(batteryItemStack, false)) {
			remoteBatteryStack = batteryItemStack.copy()
			sendNetwork(AndroidBatteryItemPacket(batteryItemStack))
		}

		if (networkQueue.size != 0) {
			for (packet in networkQueue) {
				MatteryPlayerNetworkChannel.send(ply, packet)
			}

			networkQueue.clear()
		}

		if (isAndroid != remoteIsAndroid) {
			remoteIsAndroid = isAndroid
			sendNetwork(IsAndroidPacket(isAndroid))
			shouldPlaySound = false
		}

		if (willBecomeAndroid != willBecomeAndroidNetwork) {
			willBecomeAndroidNetwork = willBecomeAndroid
			sendNetwork(WillBecomeAndroidPacket(willBecomeAndroid))
		}

		for (instance in research.values) {
			if (instance.isDirty) {
				instance.isDirty = false
				sendNetwork(AndroidResearchSyncPacket(instance))
			}
		}

		for (instance in features.values) {
			if (instance.isDirty) {
				instance.isDirty = false
				sendNetwork(AndroidFeatureSyncPacket(instance))
			}
		}
	}

	override fun extractEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		return ImpreciseFraction.ZERO
	}

	override fun extractEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		@Suppress("name_shadowing")
		var howMuch = howMuch
		var drained = ImpreciseFraction.ZERO

		if (!batteryItemStack.isEmpty) {
			batteryItemStack.getCapability(ForgeCapabilities.ENERGY).ifPresentK {
				if (it is IMatteryEnergyStorage) {
					val extracted = it.extractEnergyOuter(howMuch, simulate)
					drained += extracted
					howMuch -= extracted
				} else {
					val extracted = it.extractEnergy(howMuch, simulate)
					drained += extracted
					howMuch -= extracted
				}
			}

			if (howMuch.isZero) {
				if (!simulate && ply is ServerPlayer) {
					ply.awardStat(StatNames.POWER_CONSUMED, drained.toInt() * 10)
				}

				return drained
			}
		}

		val new = (battery - howMuch).moreThanZero()
		drained += battery - new

		if (!simulate) {
			battery = new

			if (ply is ServerPlayer) {
				ply.awardStat(StatNames.POWER_CONSUMED, drained.toInt() * 10)
			}
		}

		return drained
	}

	override fun receiveEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		@Suppress("name_shadowing")
		var howMuch = howMuch
		var received = ImpreciseFraction.ZERO

		if (!batteryItemStack.isEmpty) {
			batteryItemStack.getCapability(ForgeCapabilities.ENERGY).ifPresentK {
				if (it is IMatteryEnergyStorage) {
					val extracted = it.receiveEnergyOuter(howMuch, simulate)
					received += extracted
					howMuch -= extracted
				} else {
					val extracted = it.receiveEnergy(howMuch, simulate)
					received += extracted
					howMuch -= extracted
				}
			}

			if (howMuch.isZero) {
				return received
			}
		}

		val new = (battery + howMuch).coerceAtMost(maxBattery)
		received += new - battery

		if (!simulate) {
			battery = new
		}

		return received
	}

	override fun receiveEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		return receiveEnergyOuter(howMuch, simulate)
	}

	override var batteryLevel: ImpreciseFraction
		get() {
			if (!batteryItemStack.isEmpty) {
				batteryItemStack.getCapability(ForgeCapabilities.ENERGY).ifPresentK {
					if (it is IMatteryEnergyStorage) {
						return battery + it.batteryLevel
					} else {
						return battery + it.energyStored
					}
				}
			}

			return battery
		}
		set(value) {
			battery = value
		}

	override var maxBatteryLevel: ImpreciseFraction
		get() {
			if (batteryItemStack != ItemStack.EMPTY) {
				batteryItemStack.getCapability(ForgeCapabilities.ENERGY).ifPresentK {
					if (it is IMatteryEnergyStorage) {
						return maxBattery + it.maxBatteryLevel
					} else {
						return maxBattery + it.maxEnergyStored
					}
				}
			}

			return maxBattery
		}
		set(value) {
			maxBattery = value
		}

	private val resolver = LazyOptional.of { this }

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		return if (cap == MatteryCapability.MATTERY_PLAYER) {
			resolver.cast()
		} else LazyOptional.empty()
	}

	@Suppress("unused")
	companion object {
		@JvmField
		val UNAFFECTED_EFFECTS = ObjectArraySet<MobEffect>()

		@JvmStatic
		fun registerEffects(event: FMLCommonSetupEvent) {
			UNAFFECTED_EFFECTS.add(MobEffects.CONDUIT_POWER)
			UNAFFECTED_EFFECTS.add(MobEffects.HEAL)
			// maybe it makes things go haywire idk
			// UNAFFECTED_EFFECTS.add(MobEffects.HARM);
			UNAFFECTED_EFFECTS.add(MobEffects.REGENERATION)
			UNAFFECTED_EFFECTS.add(MobEffects.WATER_BREATHING)
			UNAFFECTED_EFFECTS.add(MobEffects.POISON)
			// even skeletons can be withered
			// UNAFFECTED_EFFECTS.add(MobEffects.WITHER);
			UNAFFECTED_EFFECTS.add(MobEffects.HEALTH_BOOST)
			UNAFFECTED_EFFECTS.add(MobEffects.ABSORPTION)
			UNAFFECTED_EFFECTS.add(MobEffects.SATURATION)
			UNAFFECTED_EFFECTS.add(MobEffects.DOLPHINS_GRACE)
		}

		@SubscribeEvent
		fun onLivingTick(event: LivingTickEvent) {
			val ent = event.entity

			if (ent.level.isClientSide) {
				ent.getCapability(MatteryCapability.MATTERY_PLAYER).ifPresentK {
					it.tickClient()
				}
			} else {
				ent.getCapability(MatteryCapability.MATTERY_PLAYER).ifPresentK {
					it.tick()
				}
			}
		}

		@SubscribeEvent(priority = EventPriority.LOWEST)
		fun onHurtEvent(event: LivingHurtEvent) {
			if (event.isCanceled) return

			event.entity.getCapability(MatteryCapability.MATTERY_PLAYER).ifPresentK { it.onHurt(event) }
		}

		const val CAPABILITY_KEY = "otm_player"
		val CAPABILITY_LOCATION = ResourceLocation(OverdriveThatMatters.MOD_ID, CAPABILITY_KEY)

		@SubscribeEvent
		fun onAttachCapabilityEvent(event: AttachCapabilitiesEvent<Entity?>) {
			val ent = event.`object`

			if (ent is Player) {
				event.addCapability(CAPABILITY_LOCATION, MatteryPlayerCapability(ent))
			}
		}

		@SubscribeEvent
		fun onPlayerChangeDimensionEvent(event: PlayerEvent.PlayerChangedDimensionEvent) {
			event.entity.getCapability(MatteryCapability.MATTERY_PLAYER)
				.ifPresentK { it.invalidateNetworkState() }
		}

		@SubscribeEvent
		fun onPlayerCloneEvent(event: PlayerEvent.Clone) {
			val it = event.entity.android ?: return

			var resolver = event.original.getCapability(MatteryCapability.MATTERY_PLAYER)

			if (!resolver.isPresent || resolver.resolve().isEmpty) {
				event.original.reviveCaps()
				resolver = event.original.getCapability(MatteryCapability.MATTERY_PLAYER)
			}

			if (!resolver.isPresent || resolver.resolve().isEmpty) {
				event.original.invalidateCaps()
				return
			}

			val original = resolver.resolve().get()

			if (original.willBecomeAndroid && event.isWasDeath) {
				original.becomeAndroid()
				(event.entity as? ServerPlayer)?.displayClientMessage(TranslatableComponent("otm.pill.message_finish").withStyle(ChatFormatting.DARK_RED), false)
			}

			it.deserializeNBT(original.serializeNBT())
			it.invalidateNetworkState()
			event.original.invalidateCaps()
		}

		val ENERGY_FOR_HUNGER_POINT = ImpreciseFraction(1000)
		const val SLEEP_TICKS_LIMIT = 80
	}
}

val ICapabilityProvider.matteryPlayer get() = getCapability(MatteryCapability.MATTERY_PLAYER).orNull()
